package cz.agents.alite.jbullet;

import javax.vecmath.Vector3f;

import com.bulletphysics.collision.broadphase.AxisSweep3;
import com.bulletphysics.collision.broadphase.BroadphaseInterface;
import com.bulletphysics.collision.dispatch.CollisionDispatcher;
import com.bulletphysics.collision.dispatch.DefaultCollisionConfiguration;
import com.bulletphysics.collision.shapes.CollisionShape;
import com.bulletphysics.dynamics.DiscreteDynamicsWorld;
import com.bulletphysics.dynamics.RigidBody;
import com.bulletphysics.dynamics.RigidBodyConstructionInfo;
import com.bulletphysics.dynamics.constraintsolver.ConstraintSolver;
import com.bulletphysics.dynamics.constraintsolver.SequentialImpulseConstraintSolver;
import com.bulletphysics.dynamics.vehicle.RaycastVehicle;
import com.bulletphysics.linearmath.DefaultMotionState;
import com.bulletphysics.linearmath.Transform;

public class BulletPhysics {

    private static BulletPhysics instance;

    private Vector3f worldMin = new Vector3f(-1500, -1500, -1500);;
    private Vector3f worldMax = new Vector3f( 1500,  1500,  1500);;
    private BroadphaseInterface overlappingPairCache;
    private ConstraintSolver constraintSolver;
    private DefaultCollisionConfiguration collisionConfiguration;
    private CollisionDispatcher dispatcher;
    private DiscreteDynamicsWorld dynamicsWorld;

    private long lastTimeStamp = 0;

    private BulletPhysics() {}

    private void initInternally(){
        //this.collisionShapes = new ArrayList<CollisionShape>();
        this.overlappingPairCache = new AxisSweep3(worldMin, worldMax, 1024);
        this.constraintSolver = new SequentialImpulseConstraintSolver();
        this.collisionConfiguration = new DefaultCollisionConfiguration();
        this.dispatcher = new CollisionDispatcher(collisionConfiguration);
        this.dynamicsWorld = new DiscreteDynamicsWorld(dispatcher, overlappingPairCache, constraintSolver, collisionConfiguration);
    }

    public void addVehicle(RaycastVehicle vehicle) {
        dynamicsWorld.addVehicle(vehicle);
    }

    public RigidBody createRigidBodyAndAddObject(float mass, CollisionShape shape, Transform startTransform) {
        // rigidbody is dynamic if and only if mass is non zero, otherwise static
        boolean isDynamic = (mass != 0f);

        Vector3f localInertia = new Vector3f(0f, 0f, 0f);
        if (isDynamic) {
            shape.calculateLocalInertia(mass, localInertia);
        }

        // using motionstate is recommended, it provides interpolation capabilities, and only synchronizes 'active' objects
        DefaultMotionState myMotionState = new DefaultMotionState(startTransform);
        RigidBodyConstructionInfo cInfo = new RigidBodyConstructionInfo(mass, myMotionState, shape, localInertia);
        RigidBody body = new RigidBody(cInfo);

        dynamicsWorld.addRigidBody(body);

        return body;
    }

    public static void init(final Vector3f worldMin, final Vector3f worldMax){
        if(instance == null){
            instance = new BulletPhysics();
            instance.worldMin = worldMin;
            instance.worldMax = worldMax;
            instance.initInternally();
        }
    }

    public static BulletPhysics getInstance(){
        if(instance == null){
            instance = new BulletPhysics();
            instance.initInternally();
        }
        return instance;
    }

    public DiscreteDynamicsWorld getDynamicsWorld() {
        return dynamicsWorld;
    }

    public void stepSimulation(long currentTime){
        //step simulation only once for the current time
        if(lastTimeStamp != currentTime){
            long stepDuration = currentTime - lastTimeStamp;
            this.dynamicsWorld.stepSimulation(stepDuration / 1000f);
            lastTimeStamp = currentTime;
        }
    }

}
