package cz.agents.alite.vis3d;

import com.jme3.app.SimpleApplication;
import com.jme3.bounding.BoundingBox;
import com.jme3.collision.CollisionResult;
import com.jme3.collision.CollisionResults;
import com.jme3.input.ChaseCamera;
import com.jme3.input.KeyInput;
import com.jme3.input.MouseInput;
import com.jme3.input.controls.ActionListener;
import com.jme3.input.controls.KeyTrigger;
import com.jme3.input.controls.MouseButtonTrigger;
import com.jme3.input.controls.Trigger;
import com.jme3.light.DirectionalLight;
import com.jme3.math.*;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;
import com.jme3.scene.Spatial;
import com.jme3.shadow.PssmShadowRenderer;
import com.jme3.shadow.PssmShadowRenderer.FilterMode;
import java.lang.Thread.UncaughtExceptionHandler;
import java.util.logging.Level;
import java.util.logging.Logger;

public class Vis3d extends SimpleApplication {

    private final UncaughtExceptionHandler defaultUncaughtExceptionHandler;

    //cam settings
    private static final float DEFAULT_NEAR =    1f;
    private static final float DEFAULT_FAR  = 1000f;
    //------------

    //flyCam settings
    private static final float CAMERA_SPEED = 50;
    private static final Trigger CAMERA_MOTION_TRIGGER = new MouseButtonTrigger(MouseInput.BUTTON_RIGHT);
    private Vector3f defaultCamLocation;
    private Vector3f defaultCamLookAt;
    //---------------

    //world node settings
    private Transform worldTransform = new Transform();
    //-------------------

    private Vis3dManager manager;

    private Node world;
    private Node chaseable;

    private PssmShadowRenderer pssmRenderer;

    private boolean chasing = false;
    private Spatial chasedSpatial;

    private Vector3f lastPos;

    private ChaseCamera chaseCam;

    protected Vis3d(Vis3dManager manager) {
        Logger.getLogger("com.jme3").setLevel(Level.WARNING);
        defaultUncaughtExceptionHandler = Thread.getDefaultUncaughtExceptionHandler();
        this.manager = manager;
    }

    @Override
    public void simpleInitApp() {
        pssmRenderer = new PssmShadowRenderer(getAssetManager(), 1024, 1);
        pssmRenderer.initialize(renderManager, viewPort);
        pssmRenderer.setFilterMode(FilterMode.PCF8);
        pssmRenderer.setLambda(0.5f);
        viewPort.getProcessors().clear();
        viewPort.addProcessor(pssmRenderer);

        initCameraSettings();

        initFlyCameraSettings();

        //call users settings
        manager.initSettingLayer(this);
        //-------------------

        //world
        createWorld();
        //-------

        //chaseable
        chaseable = new Node("pick node");
        world.attachChild(chaseable);
        //----

        //register environment entities
        manager.initLayers(this);
        //-----------------------------

        initLights();

        registerInput();

        resetCamera();

        // bypass of jME3 ?bug?
        Thread.setDefaultUncaughtExceptionHandler(defaultUncaughtExceptionHandler);
    }

    public Vector2f getDisplayDimension(){
        return new Vector2f(settings.getWidth(), settings.getHeight());
    }

    @Override
    public void loadStatsView() {
    }

    private void initCameraSettings() {
        cam.setFrustumNear(DEFAULT_NEAR);
        cam.setFrustumFar(DEFAULT_FAR);
    }

    private void initLights() {
        DirectionalLight dl1 = new DirectionalLight();
        DirectionalLight dl2 = new DirectionalLight();
        DirectionalLight dl3 = new DirectionalLight();
        dl1.setDirection(new Vector3f(-1, -1, -1));
        dl1.setColor(new ColorRGBA(1, 1, .85f, 1));
        dl2.setDirection(new Vector3f(1, -1, 1));
        dl2.setColor(ColorRGBA.Gray);
        dl3.setDirection(new Vector3f(0, -10, 0));
        dl3.setColor(new ColorRGBA(.5f, .5f, .425f, 1));

        rootNode.addLight(dl1);
        rootNode.addLight(dl2);
        rootNode.addLight(dl3);
    }

    private void initFlyCameraSettings() {
        inputManager.deleteTrigger("FLYCAM_RotateDrag", new MouseButtonTrigger(MouseInput.BUTTON_LEFT));
        inputManager.addMapping("FLYCAM_RotateDrag", CAMERA_MOTION_TRIGGER);
        flyCam.setMoveSpeed(CAMERA_SPEED);
        flyCam.setDragToRotate(true);
    }

    private void registerInput(){
        inputManager.addMapping("Select", new MouseButtonTrigger(0));
        inputManager.addListener(selectionListener, "Select");

        inputManager.addMapping("resetCam", new KeyTrigger(KeyInput.KEY_R));
        inputManager.addListener(resetCamToggle, "resetCam");
    }

    private void createWorld() {
        world = new Node("world");
        world.setLocalTransform(worldTransform);
        world.setModelBound(new BoundingBox());
        world.updateModelBound();
        rootNode.attachChild(world);
    }

    public void registerToChaseable(Spatial s) {
        chaseable.attachChild(s);
    }

    public void registerToRootNode(Spatial s){
        rootNode.attachChild(s);
    }

    public void registerToWorld(Spatial s){
        world.attachChild(s);
    }

    public void registerToGuiNode(Spatial s){
        guiNode.attachChild(s);
    }

    public void setWorldTransform(Transform worldTransform) {
        this.worldTransform = worldTransform;
    }

    public void setDefaultCamLocation(Vector3f defaultCamLocation) {
        this.defaultCamLocation = defaultCamLocation;
    }

    public void setDefaultCamLookAt(Vector3f defaultCamLookAt) {
        this.defaultCamLookAt = defaultCamLookAt;
    }

    public Spatial getChasedSpatial() {
        return chasedSpatial;
    }

    private ActionListener selectionListener = new ActionListener() {
        @Override
        public void onAction(String name, boolean keyPressed, float tpf) {
            if (name.equals("Select") && !keyPressed) {

                //get mouse coordinates and compute relative direction
                Vector2f cursorPosition = inputManager.getCursorPosition();
                Vector3f pos = cam.getWorldCoordinates(cursorPosition, 0.0f);
                Vector3f dir = cam.getWorldCoordinates(cursorPosition, 1f);
                dir.subtractLocal(pos).normalizeLocal();

                //process ray and chase chosen collidables
                Ray r = new Ray(pos, dir);
                CollisionResults res = new CollisionResults();
                int num = chaseable.collideWith(r, res);
                CollisionResult closest;
                if (num > 0) {
                    closest = res.getClosestCollision();
                    if (closest != null){
                        Geometry g = closest.getGeometry();
                        chase(g);
                    }
                } else {
                    stopChase();
                }
            }
        }
    };

    private ActionListener resetCamToggle = new ActionListener() {
        @Override
        public void onAction(String name, boolean isPressed, float tpf) {
            if (name.equals("resetCam") && !isPressed) {
                resetCamera();
            }
        }
    };

    private void chase(Spatial chaseable) {
        if (!chasing) {
            chasing = true;
            chasedSpatial = chaseable;
            lastPos = cam.getLocation();
            flyCam.setEnabled(false);

            //hack z axis align
            cam.setAxes(Vector3f.UNIT_X, Vector3f.UNIT_Y, Vector3f.UNIT_Z);
            chaseCam = new ChaseCamera(cam, chaseable, inputManager);
            chaseCam.setInvertVerticalAxis(true);
//				chaseCam.setMaxDistance(3);
            chaseCam.setMinDistance(2f);
            chaseCam.setChasingSensitivity(1);
//				chaseCam.setDefaultHorizontalRotation(0);
//				chaseCam.setDefaultVerticalRotation(0);
            chaseCam.setSmoothMotion(true);
        }
    }

    protected void stopChase() {
        if(chasing){
            chasing = false;
            chasedSpatial = null;
            chaseCam.setEnabled(false);
            chaseCam = null;
            cam.setLocation(lastPos);
            cam.update();
            flyCam.setEnabled(true);
        }
    }

    protected void resetCamera(){
        stopChase();
        cam.setLocation(defaultCamLocation);
        cam.lookAt(defaultCamLookAt, Vector3f.UNIT_Y);
    }

    @Override
    public void simpleUpdate(float tpf) {
        manager.initPostponedLayers();

        super.simpleUpdate(tpf);
    }

}
