package tacv.universe.world;

import com.bulletphysics.collision.shapes.BvhTriangleMeshShape;
import com.bulletphysics.collision.shapes.TriangleIndexVertexArray;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.List;
import tacv.universe.world.xml.*;

public class CollisionMeshLoader {

    public static List<BvhTriangleMeshShape> loadCollisionMesh(CollisionMesh xmlCollision) {
        List<BvhTriangleMeshShape> shapes = new LinkedList<BvhTriangleMeshShape>();


        Heightmap xmlHeightmap = xmlCollision.getHeightmap();
        BvhTriangleMeshShape heightmapShape = loadHeightmap(xmlHeightmap);
        shapes.add(heightmapShape);

        for (Box xmlBox : xmlCollision.getBoxes().getBox()) {
            BvhTriangleMeshShape boxShape = loadBox(xmlBox);
            if (boxShape != null) {
                shapes.add(boxShape);
            }
        }

        for (Mesh xmlMesh : xmlCollision.getMeshes().getMesh()) {
            BvhTriangleMeshShape meshShape = loadMesh(xmlMesh);
            shapes.add(meshShape);
        }

        return shapes;
    }

    private static BvhTriangleMeshShape loadHeightmap(Heightmap xmlHeightmap) {
        double[][] heightmapData = new double[xmlHeightmap.getArraySizeX()][xmlHeightmap.getArraySizeY()];
        for (Data data : xmlHeightmap.getData()) {
            heightmapData[data.getI()][data.getJ()] = data.getAltitude();
        }

        int vertexSize = 4 * 3;
        int indexSize = 4;
        int faceCountX = xmlHeightmap.getArraySizeX() - 1;
        int faceCountY = xmlHeightmap.getArraySizeY() - 1;
        double faceSizeX = xmlHeightmap.getTerrainSizeX() / faceCountX;
        double faceSizeY = xmlHeightmap.getTerrainSizeY() / faceCountY;
        int vertexCount = xmlHeightmap.getArraySizeX() * xmlHeightmap.getArraySizeY();
        int totalTriangles = faceCountX * faceCountY * 2;

        ByteBuffer vertices = ByteBuffer.allocateDirect(vertexCount * vertexSize).order(ByteOrder.nativeOrder());
        ByteBuffer indices = ByteBuffer.allocateDirect(totalTriangles * 3 * indexSize).order(ByteOrder.nativeOrder());

        double y = xmlHeightmap.getTerrainOriginY();
        for (int j = 0; j < xmlHeightmap.getArraySizeY(); j++) {
            double x = xmlHeightmap.getTerrainOriginX();
            for (int i = 0; i < xmlHeightmap.getArraySizeX(); i++) {
                vertices.putFloat((float) x);
                vertices.putFloat((float) heightmapData[i][j]);
                vertices.putFloat((float) -y);

                //add to height map info
                x += faceSizeX;
            }
            y += faceSizeY;
        }

        for (int j = 0; j < faceCountY; j++) {
            int baseIndex = j * (faceCountX + 1);
            for (int i = 0; i < faceCountX; i++) {
                indices.putInt(baseIndex);
                indices.putInt(baseIndex + 1);
                indices.putInt(baseIndex + faceCountX + 1);
                indices.putInt(baseIndex + 1);
                indices.putInt(baseIndex + faceCountX + 2);
                indices.putInt(baseIndex + faceCountX + 1);
                baseIndex++;
            }
        }

        TriangleIndexVertexArray indexVertexArrays =
                new TriangleIndexVertexArray(totalTriangles, indices, indexSize * 3, vertexCount, vertices, vertexSize);
        return new BvhTriangleMeshShape(indexVertexArrays, true);
    }

    private static BvhTriangleMeshShape loadBox(Box xmlBox) {
        int pointCount = xmlBox.getPoint().size();
        if (pointCount < 3) {
            return null;
        }
        Point[] xmlPoints = new Point[pointCount];
        xmlBox.getPoint().toArray(xmlPoints);

        int vertexSize = 4 * 3;
        int indexSize = 4;
        int vertexCount = pointCount * 2;
        int totalTriangles = (pointCount + 2) * 2;

        ByteBuffer vertices = ByteBuffer.allocateDirect(vertexCount * vertexSize).order(ByteOrder.nativeOrder());
        ByteBuffer indices = ByteBuffer.allocateDirect(totalTriangles * 3 * indexSize).order(ByteOrder.nativeOrder());

        for (int i = 0; i < pointCount; i++) {
            vertices.putFloat((float) xmlPoints[i].getX());
            vertices.putFloat((float) xmlBox.getBaseAltitude());
            vertices.putFloat((float) -xmlPoints[i].getY());
        }
        for (int i = 0; i < pointCount; i++) {
            vertices.putFloat((float) xmlPoints[i].getX());
            vertices.putFloat((float) (xmlBox.getBaseAltitude() + xmlBox.getHeight()));
            vertices.putFloat((float) -xmlPoints[i].getY());
        }

        for (int i = 0; i < pointCount; i++) {
            int i1 = i;
            int i2 = (i1 + 1) % pointCount;
            indices.putInt(i1);
            indices.putInt(i2);
            indices.putInt(i2 + pointCount);
            indices.putInt(i1);
            indices.putInt(i2 + pointCount);
            indices.putInt(i1 + pointCount);
        }

        for (int i = 0; i < pointCount - 2; i++) {
            indices.putInt(0);
            indices.putInt(i + 1);
            indices.putInt(i + 2);
        }
        for (int i = 0; i < pointCount - 2; i++) {
            indices.putInt(pointCount);
            indices.putInt(pointCount + i + 1);
            indices.putInt(pointCount + i + 2);
        }

        TriangleIndexVertexArray indexVertexArrays =
                new TriangleIndexVertexArray(totalTriangles, indices, indexSize * 3, vertexCount, vertices, vertexSize);
        return new BvhTriangleMeshShape(indexVertexArrays, true);
    }

    private static BvhTriangleMeshShape loadMesh(Mesh xmlMesh) {
        int vertexSize = 4 * 3;
        int indexSize = 4;
        int vertexCount = xmlMesh.getVertices().getVertex().size();
        int faceCount = xmlMesh.getFaces().getFace().size();

        Vertex[] xmlVertices = new Vertex[vertexCount];
        for (Vertex vertex : xmlMesh.getVertices().getVertex()) {
            xmlVertices[vertex.getId()] = vertex;
        }

        ByteBuffer vertices = ByteBuffer.allocateDirect(vertexCount * vertexSize).order(ByteOrder.nativeOrder());
        ByteBuffer indices = ByteBuffer.allocateDirect(faceCount * 3 * indexSize).order(ByteOrder.nativeOrder());

        for (Vertex vertex : xmlVertices) {
            vertices.putFloat((float) vertex.getX());
            vertices.putFloat((float) vertex.getZ());
            vertices.putFloat((float) -vertex.getY());
        }

        for (Face face : xmlMesh.getFaces().getFace()) {
            indices.putInt(face.getVertexID1());
            indices.putInt(face.getVertexID2());
            indices.putInt(face.getVertexID3());
        }

        TriangleIndexVertexArray indexVertexArrays =
                new TriangleIndexVertexArray(faceCount, indices, indexSize * 3, vertexCount, vertices, vertexSize);
        return new BvhTriangleMeshShape(indexVertexArrays, true);
    }
}
