package tacv.universe.world;

import java.util.ArrayList;
import java.util.List;

import javax.vecmath.Point2f;
import javax.vecmath.Point3f;

import tacv.universe.world.xml.CollisionMesh;
import tacv.universe.world.xml.Data;
import tacv.universe.world.xml.Heightmap;

public class Terrain {

    private float worldSizeX;
    private float worldSizeY;
    private int cellCountX;
    private int cellCountY;
    private float cellSizeX;
    private float cellSizeY;
    private float[][] heightmap;

    private TriangleInterpolator tri1, tri2;
    private List<Face> extraFaces = new ArrayList<Face>();

    private float[] coefs = new float[3];


    private Terrain(CollisionMesh xmlCollision) {
        init(xmlCollision);
    }

    public static Terrain create(CollisionMesh xmlCollision){
        return new Terrain(xmlCollision);
    }

    private void init(CollisionMesh xmlCollision){
        Heightmap xmlHeightmap = xmlCollision.getHeightmap();

        this.worldSizeX = (float)xmlHeightmap.getTerrainSizeX();
        this.worldSizeY = (float)xmlHeightmap.getTerrainSizeY();
        cellCountX = xmlHeightmap.getArraySizeX();
        cellCountY = xmlHeightmap.getArraySizeY();
        cellSizeX = this.worldSizeX / (xmlHeightmap.getArraySizeX() - 1);
        cellSizeY = this.worldSizeY / (xmlHeightmap.getArraySizeY() - 1);

        heightmap = new float[cellCountX][cellCountY];
        for (Data data : xmlHeightmap.getData()) {
            heightmap[data.getI()][data.getJ()] = (float) data.getAltitude();
        }

        tri1 = new TriangleInterpolator(new Point2f(0, 0), new Point2f(1, 0), new Point2f(0, 1));
        tri2 = new TriangleInterpolator(new Point2f(1, 1), new Point2f(0, 1), new Point2f(1, 0));

        defineExtraFaces();
    }

    private void defineExtraFaces() {
        Point3f v1 = new Point3f(382.192169f, 285.146179f, 33.849575f);
        Point3f v2 = new Point3f(399.295349f, 274.782410f, 33.578175f);
        Point3f v3 = new Point3f(423.628082f, 353.396912f, 38.840153f);
        Point3f v4 = new Point3f(440.731262f, 343.033142f, 38.568752f);
        extraFaces.add(new Face(v1, v2, v4));
        extraFaces.add(new Face(v1, v4, v3));
    }

    public float getAltitude(float x, float y) {
        int i = (int) Math.floor(x / cellSizeX);
        int j = (int) Math.floor(y / cellSizeY);
        if (i < 0 || i > cellCountX - 2 ||
            j < 0 || j > cellCountY - 2) {
            return 0;
        }
        float xx = x / cellSizeX - i;
        float yy = y / cellSizeY - j;
        float altitude = 0;
        if (tri1.calcContribs(xx, yy, coefs)) {
            altitude = heightmap[i][j] * coefs[0] + heightmap[i + 1][j] * coefs[1] + heightmap[i][j + 1] * coefs[2];
        }
        if (tri2.calcContribs(xx, yy, coefs)) {
            altitude = heightmap[i + 1][j + 1] * coefs[0] + heightmap[i][j + 1] * coefs[1] + heightmap[i + 1][j] * coefs[2];
        }
        for (Face face : extraFaces) {
            Float alt = face.getAltitude(x, y);
            if (alt != null && alt > altitude) {
                altitude = alt;
            }
        }
        return altitude;
    }

    private static class TriangleInterpolator {

        private Point2f v1;
        private Point2f v2;
        private Point2f v3;

        private float xMin, yMin, xMax, yMax;


        public TriangleInterpolator(Point2f v1, Point2f v2, Point2f v3) {
            this.v1 = v1;
            this.v2 = v2;
            this.v3 = v3;
            xMin = Math.min(Math.min(v1.x, v2.x), v3.x);
            yMin = Math.min(Math.min(v1.y, v2.y), v3.y);
            xMax = Math.max(Math.max(v1.x, v2.x), v3.x);
            yMax = Math.max(Math.max(v1.y, v2.y), v3.y);
        }

        public boolean isInsideBoundingBox(float x, float y) {
            if (x < xMin || x > xMax || y < yMin || y > yMax) {
                return false;
            }
            return true;
        }

        public boolean calcContribs(float x, float y, float[] coefs) {
            double ux = x - v1.x;
            double uy = y - v1.y;
            if (ux == 0 && uy == 0) {
                coefs[0] = 1.0f;
                coefs[1] = 0.0f;
                coefs[2] = 0.0f;
                return true;
            }
            double vx = v3.x - v2.x;
            double vy = v3.y - v2.y;
            double wx = v2.x - v1.x;
            double wy = v2.y - v1.y;
            double m1 = uy * vx - ux * vy;
            if (m1 == 0) {
                coefs[0] = 0.0f;
                coefs[1] = 0.0f;
                coefs[2] = 0.0f;
                return false;
            }
            double m2 = vx * wy - vy * wx;
            if (m2 == 0) {
                coefs[0] = 0.0f;
                coefs[1] = 0.0f;
                coefs[2] = 0.0f;
                return false;
            }
            double s = m1 / m2;
            if (s < 0 || s > 1) {
                coefs[0] = 0.0f;
                coefs[1] = 0.0f;
                coefs[2] = 0.0f;
                return false;
            }
            double m3 = ux * wy - uy * wx;
            double t = m3 / m1;
            if (t < 0 || t > 1) {
                coefs[0] = 0.0f;
                coefs[1] = 0.0f;
                coefs[2] = 0.0f;
                return false;
            }
            coefs[0] = (float) (1 - s);
            coefs[1] = (float) ((1 - t) * s);
            coefs[2] = (float) (t * s);
            return true;
        }
    }

    private static class Face {

        private TriangleInterpolator tri;
        private float[] z;
        private float[] coefs = new float[3];

        public Face(Point3f v1, Point3f v2, Point3f v3) {
            tri = new TriangleInterpolator(
                    new Point2f(v1.x, v1.y),
                    new Point2f(v2.x, v2.y),
                    new Point2f(v3.x, v3.y));
            z = new float[]{v1.z, v2.z, v3.z};
        }

        public Float getAltitude(float x, float y) {
            if (!tri.isInsideBoundingBox(x, y)) {
                return null;
            }
            if (!tri.calcContribs(x, y, coefs)) {
                return null;
            }
            return z[0] * coefs[0] + z[1] * coefs[1] + z[2] * coefs[2];
        }
    }

}
